#include "Game.h"
#include <random>

Game::Game(Difficulty difficulty) :_pView(new View(*this))
{
	SetDifficulty(difficulty);
}

Game::~Game()
{
}


void Game::Show()
{
	_pView->ShowGame();
}

void Game::SetDifficulty(Difficulty newDifficulty)
{
	_Difficulty = newDifficulty;
	switch (newDifficulty)
	{
	case easy:
		for (int i = 0; i < GroupsNum; ++i)
		{
			for (int j = 0; j < CardValueNum; ++j)
			{
				_Cards[i * CardValueNum + j] = Card((CardValue)j, CardType::Spade);
			}
		}
		break;
	case normal:
		for (int i = 0; i < GroupsNum; ++i)
		{
			for (int j = 0; j < CardValueNum; ++j)
			{
				_Cards[i * CardValueNum + j] = Card((CardValue)j, j % 2 ? CardType::Heart : CardType::Spade);
			}
		}
		break;
	case hard:
		for (int i = 0; i < GroupsNum; ++i)
		{
			for (int j = 0, k = 0; j < CardValueNum; ++j, k = (k + 1) % 4)
			{
				_Cards[i * CardValueNum + j] = Card((CardValue)j, (CardType)k);
			}
		}
		break;
	default:
		break;
	}
}


void Game::Shuffle()
{
 	static std::random_device rd;
	std::default_random_engine e1(rd());
	std::uniform_int_distribution<int> uniform_dist(0, _Cards.size() - 1);
	for (int i = 0; i < _Cards.size(); ++i)
	{
		std::swap(_Cards[i], _Cards[uniform_dist(e1)]);
	}
}

void Game::InitScenes()
{
	ReservedCount = ReservedQuantity;
	SuccessCount = 0;
	IsVictory = false;
	for (int i = 0; i < Scenes.size(); ++i)
		Scenes[i].Clear();

	for (int i = ReservedQuantity * Scenes.size(), j = 0; i < _Cards.size(); ++i, j = (j+1) % Scenes.size())
		Scenes[j].Push(_Cards[i]);
	
	for (int i = 0; i < Scenes.size(); ++i)
		Scenes[i].SetHideLevel(Scenes[i].Count() - 1);
	
	PSelected.SetValue(-1, -1);
	PCurrent.SetValue(0, Scenes[0].Count() - 1);
}

void Game::StartNewGame()
{
	Shuffle();
	InitScenes();
}

void Game::Deal()
{
	if (ReservedCount < 1)
		return;

	// ֮ǰѡеȡ
	PSelected.SetValue(-1, -1);

	// Ϊÿһ۷һ
	for (int i = 0; i < Scenes.size(); ++i)
	{
		Scenes[i].Push(_Cards[(ReservedCount-1) * Scenes.size() + i]);
	}
	// һ
	ReservedCount--;
}

void Game::CurMove(Dir dir)
{
	int width = Scenes.size();
	int height = Scenes[PCurrent.x].Count();
	switch (dir)
	{
	case left:
		if (PCurrent.x == 0)
			PCurrent.x = width -1;
		else
			PCurrent.x--;
		break;
	case up:
		if (height == 0)
			PCurrent.y = 0;
		else if (PCurrent.y == 0)
			PCurrent.y = height - 1;
		else
			PCurrent.y--;
		break;
	case right:
		if (PCurrent.x == width - 1)
			PCurrent.x = 0;
		else
			PCurrent.x++;
		break;
	case down:
		if (height == 0)
			PCurrent.y = 0;
		else if (PCurrent.y == height - 1)
			PCurrent.y = 0;
		else
			PCurrent.y++;
		break;
	default:
		break;
	}

	// һûκ  ָ
	if (Scenes[PCurrent.x].Count() == 0)
		PCurrent.y = 0;
	// ߶ȸڵǰ߶
	//     ֮ǰĸ߶0
	//     ֮ǰһеĶʱ
	//  ָǰеĶ
	else if (PCurrent.y >= Scenes[PCurrent.x].Count()
		|| height == 0
		|| PCurrent.y == height - 1)
		PCurrent.y = Scenes[PCurrent.x].Count() - 1;
}

bool Game::Select()
{
	if (PCurrent.x < 0 || PCurrent.x >= Scenes.size())
		return false;

	if (Scenes[PCurrent.x].Count() == 0 && PCurrent.y != 0)
		return false;
	
	// ѡ֮ǰѡģȡѡ
	if (PSelected == PCurrent)
	{
		PSelected.SetValue(-1, -1);
		return true;
	}

	// ֮ǰδѡκζ
	if (PSelected.x == -1)
	{
		// жϵǰѡɲѡ
		if (CheckIsOptional(PCurrent))
		{
			// ѡѡλãtrue
			PSelected = PCurrent;
			return true;
		}
		else
		{
			return false;
		}
	}
	// ˵֮ǰѡ˶ƶ
	else
	{
		// ѡĺ֮ǰѡͬһ  ѡǲм
		if (PCurrent.x == PSelected.x || PCurrent.y != Scenes[PCurrent.x].Count() - 1)
		{
			// һɶû
			if (Scenes[PCurrent.x].Count() == 0)
			{
				// ֱƶȥ
				MoveCards(PSelected, PCurrent.x); 
				PSelected.SetValue(-1, -1);
				return true;
			}
			if (CheckIsOptional(PCurrent))
			{
				PSelected = PCurrent;
				return true;
			}
			else
			{
				return false;
			}
		}
		// ˵ѡеһ
		else
		{
			// ƶȥ
			if (CheckIsOrderly(
				Scenes[PCurrent.x].GetCards().at(PCurrent.y),
				Scenes[PSelected.x].GetCards().at(PSelected.y),
				_Difficulty))
			{
				// ֱƶȥ
				MoveCards(PSelected, PCurrent.x);
				PSelected.SetValue(-1, -1);
				return true;
			}
			// ѡһ
			else
			{
				PSelected = PCurrent;
				return true;
			}
		}
	}
}

void Game::MoveCards(Point point, int SlotsIndex)
{
	// ָλõƶӦĲ
	Scenes[point.x].MoveCardsTo(Scenes[SlotsIndex], point.y);

	// Ƿ
	if (CheckIsSuccess(SlotsIndex))
	{
		// һ飬򵯳һ
		Scenes[SlotsIndex].Pop(CardValueNum);
		// Ȼɼ
		SuccessCount++;

		if (SuccessCount == GroupsNum)
			IsVictory = true;
	}
}


bool Game::CheckIsOptional(Point point)
{
	// һɶû
	if (Scenes[point.x].Count() == 0)
		return false;

	if (point.y < Scenes[point.x].GetHideLevel())
		return false;

	if (point.y == Scenes[point.x].Count() - 1)
		return true;

	
	for (int i = point.y; i < Scenes[point.x].Count() - 1; ++i)
	{
		// ǷֻҪһǴģֱ޷ѡ
		if (!CheckIsOrderly(Scenes[point.x].GetCards().at(i), Scenes[point.x].GetCards().at(i + 1), _Difficulty))
		{
			return false;
		}
	}

	return true;
}

bool Game::CheckIsOrderly(Card cur, Card next, Difficulty difficulty)
{
	if (cur.GetValue() != next.GetValue() + 1)
		return false;

	// Ѷȼ黨ɫ
	switch (difficulty)
	{
	// ѶȲɫ
	case easy:
		return true;

	// ͨѶҪǷں
	case normal:
		return cur.GetType() == CardType::Spade && next.GetType() == CardType::Heart
			|| cur.GetType() == CardType::Heart && next.GetType() == CardType::Spade;
	
	// ѶҪǷպҡҡ÷˳
	case hard:
		return cur.GetType() == CardType::Spade		&& next.GetType() == CardType::Heart
			|| cur.GetType() == CardType::Heart		&& next.GetType() == CardType::Club
			|| cur.GetType() == CardType::Club		&& next.GetType() == CardType::Diamond
			|| cur.GetType() == CardType::Diamond	&& next.GetType() == CardType::Spade;
	default:
		break;
	}
	return false;
}

// һǷ
bool Game::CheckIsSuccess(int SlotsIndex)
{
	// ж۵㲻㹻һ
	if (Scenes[SlotsIndex].Count() < CardValueNum)
		return false;

	// ʾһƣֱӷ
	if (Scenes[SlotsIndex].Count() - Scenes[SlotsIndex].GetHideLevel() < CardValueNum)
		return false;

	std::vector<Card> Cards = Scenes[SlotsIndex].GetCards();
	// ȻжǲAֱӷ
	if (Cards.back().GetValue() != CardValue::_A)
		return false;

	// Ȼ鵹һλÿʼǷֱӷؼ
	return CheckIsOptional(Point(SlotsIndex, Cards.size() - CardValueNum));
}
